/* Select-based JOH loop.
   Copyright (C) 2011 Sergey Poznyakoff

   JOH is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   JOH is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JOH.  If not, see <http://www.gnu.org/licenses/>. */

#include "joh.h"

struct kwdef {
	const char *kws;
	size_t kwl;
	size_t param;
};

static struct kwdef kwtab[] = {
	{ "id",     2, joh_param_id },
	{ "key",    3, joh_param_key },
	{ "newkey", 6, joh_param_newkey },
	{ "body",   4, joh_param_body },
	{ "data",   4, joh_param_body },
	{ "server", 6, joh_param_server },
	{ "port",   4, joh_param_port },
	{ NULL }
};

static struct kwdef *
find_kw(const char *field, size_t len)
{
	struct kwdef *p;

	for (p = kwtab; p->kws; p++)
		if (p->kwl == len && memcmp(p->kws, field, len) == 0)
			return p;
	return NULL;
}

int
client_decode(struct joh_param param[joh_param_max],
	      joh_cbuf_t ibuf, const char *id)
{
	char *data, *ptr;
	uint32_t size;
	size_t cbufsize = joh_cbuf_size(ibuf);
	unsigned char i, nelem;
	
	if (cbufsize < sizeof(size))
		return -1;
	size = ntohl(*(uint32_t*) joh_cbuf_head(ibuf, NULL));
	if (cbufsize - sizeof(size) < size)
		return -1;
	data = joh_cbuf_finish(ibuf);
	data += sizeof(size);
	nelem = *data++;
	for (i = 0, ptr = data; i < nelem && ptr < data + size; i++) {
		struct kwdef *kw;
		uint16_t len;

		len = ntohs(*(unsigned short *)ptr);
		ptr += sizeof(len);

		kw = find_kw(ptr, len);
		if (!kw)
			warn("ignoring unknown keyword %*.*s",
			     (int) len, (int) len, ptr);

		ptr += len;

		len = ntohs(*(unsigned short *)ptr);
		ptr += sizeof(len);

		if (kw) {
			char *p = xmalloc(len + 1);
			memcpy(p, ptr, len);
			p[len] = 0;
			param[kw->param].ptr = p;
			param[kw->param].size = len;
			debug(50, ("%s: %s=%s", id, kw->kws, p));

		}
		ptr += len;
	}
	return 0;
}

int
client_decode_session(struct joh_session *sess, struct joh_sockaddr *sa)
{
	int rc = client_decode(sess->param, sess->ibuf, joh_sockaddr_str(sa));
	if (rc == 0) {
		if (!sess->param[joh_param_id].ptr) {
			error("no ID in packet from %s", joh_sockaddr_str(sa));
			return -1;
		}
		if (!sess->param[joh_param_key].ptr) {
			error("no key in packet from %s", joh_sockaddr_str(sa));
			return -1;
		}
	}
	return rc;
}

int
client_encode(struct joh_obuf *buf, size_t count, const char **param_names,
	      struct joh_param *param_vals)
{
	char *ptr;
	size_t total, num;
	size_t i;
	
	total = sizeof(uint32_t) + 1;
	num = 0;
	for (i = 0; i < count; i++)
		if (param_vals[i].size) {
			num++;
			total += strlen(param_names[i]) + param_vals[i].size +
				  2 * sizeof(uint16_t);
		}
		
	buf->base = xmalloc(total);
	buf->base_alloc = 1;
	buf->size = total;
	buf->off = 0;

	ptr = buf->base;
	*(uint32_t*)ptr = htonl(total - sizeof(uint32_t));
	ptr += sizeof(uint32_t);
	*ptr++ = (unsigned char) num;

	for (i = 0; i < count; i++) {
		size_t param_len;

		if (param_vals[i].size == 0)
			continue;
		param_len = strlen(param_names[i]);
		*(uint16_t*)ptr = htons(param_len);
		ptr += sizeof(uint16_t);
		memcpy(ptr, param_names[i], param_len);
		ptr += param_len;

		*(uint16_t*)ptr = htons(param_vals[i].size);
		ptr += sizeof(uint16_t);
		memcpy(ptr, param_vals[i].ptr, param_vals[i].size);
		ptr += param_vals[i].size;
	}
	return 0;
}
	
